/*============================================================================
  File:     CreditPV-Setup.sql

  Summary:  This script has all of the syntax needed to create the a partitioned
			table structure to compare/contrast partitioned views and partitioned
			tables within the Credit database.
			
			This will allow you to see the differences between PVs and PTs
			using two comparable partitioning stategies.

  Date:     December 2010

  SQL Server Version: SQL Server 2005/2008
------------------------------------------------------------------------------
  Written by Kimberly L. Tripp & Paul S. Randal, SQLskills.com
  All rights reserved.

  For more scripts and sample code, check out http://www.SQLskills.com

  This script is intended only as a supplement to demos and lectures
  given by SQLskills.com  
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

-- Walk your way through this script... there's a TON of cool stuff in here. 
-- (1) You'll setup for Partitioned Views by creating four files and four filegroups 
--     for the Partitioned View and a last file/filegroup for staging data. 
-- (2) You'll copy data into a staging area and then build the clustered indexes
--     on the appropriate filegroup.
-- (3) Once all of the data is partitioned then you'll create a partitioned view using 
--     UNION ALL.
-- (4) You'll run queries against the Partitioned View and you'll see how the optimizer
--     removes redundant data from the execution plan. 
-- (5) What about indexed views used with partitioned views. No, you can't index a partitioned
--     view however, you can create views and then index them - against the base tables.
--     Take for example the sum of charge by member. If you've partitioned June/July/Aug, etc.
--     then summing charges means summing the sums (from the partitions). Is the optimizer
--     smart enough... I guess you'll have to get to the end of the script to see ;).

-- Have fun,
-- kt

USE master
GO

/* ******* pre-step *******  */
RESTORE DATABASE CreditPV
          FROM DISK = 'C:\SQLskills\CreditBackup100.BAK'
WITH MOVE N'CreditData' TO N'C:\Program Files\Microsoft SQL Server\MSSQL10.SQLDev01\mssql\data\CreditPVData.mdf'
        , MOVE N'CreditLog' TO N'C:\Program Files\Microsoft SQL Server\MSSQL10.SQLDev01\mssql\data\CreditPVLog.ldf'
        , REPLACE
go
/* ******* pre-step ******* */

USE CreditPV
go

ALTER DATABASE CreditPV 
	ADD FILEGROUP PartitionedViewsFilegroup1
go

ALTER DATABASE CreditPV
	ADD FILE
		(NAME = N'CreditPVFGFile1', 
		FILENAME = N'C:\SQLskills\CreditPVFGFile1.NDF' , 
		SIZE = 50, FILEGROWTH = 10%) 
		TO FILEGROUP [PartitionedViewsFilegroup1]
go

ALTER DATABASE CreditPV 
	ADD FILEGROUP PartitionedViewsFilegroup2
go

ALTER DATABASE CreditPV
	ADD FILE
		(NAME = N'CreditPVFGFile2', 
		FILENAME = N'C:\SQLskills\CreditPVFGFile2.NDF' , 
		SIZE = 50, FILEGROWTH = 10%) 
		TO FILEGROUP [PartitionedViewsFilegroup2]
go

ALTER DATABASE CreditPV
	ADD FILEGROUP PartitionedViewsFilegroup3
go

ALTER DATABASE CreditPV
	ADD FILE
		(NAME = N'CreditPVFGFile3', 
		FILENAME = N'C:\SQLskills\CreditPVFGFile3.NDF' , 
		SIZE = 50, FILEGROWTH = 10%) 
		TO FILEGROUP [PartitionedViewsFilegroup3]
go

ALTER DATABASE CreditPV
	ADD FILEGROUP PartitionedViewsFilegroup4
go

ALTER DATABASE CreditPV
	ADD FILE
		(NAME = N'CreditPVFGFile4', 
		FILENAME = N'C:\SQLskills\CreditPVFGFile4.NDF' , 
		SIZE = 50, FILEGROWTH = 10%) 
		TO FILEGROUP [PartitionedViewsFilegroup4]
go

sp_helpfile
-- Move a "staging" copy into it's own filegroup...
ALTER DATABASE CreditPV
	ADD FILEGROUP ChargesStaging
go

ALTER DATABASE CreditPV
	ADD FILE
		(NAME = N'ChargesStaging', 
		FILENAME = N'C:\SQLskills\CreditPVChargesStaging.NDF' , 
		SIZE = 100, FILEGROWTH = 10%) 
		TO FILEGROUP [ChargesStaging]
go

-- Load the data into the Staging Area and then rebuild the clustered index on the filegroup to move
-- the object once and to get better contiguousness...

----------------------------------------------------------------------------------------------
--------------------------------------- JUNE --------------------------------------------
----------------------------------------------------------------------------------------------
CREATE TABLE Charges200406
( 	charge_no		numeric_id		NOT NULL	IDENTITY,
	member_no		numeric_id 	NOT NULL
						CONSTRAINT Charges200406MemberNoFK
							REFERENCES dbo.Member(Member_No),
	provider_no		numeric_id
						CONSTRAINT Charges200406ProviderNoFK
							REFERENCES dbo.Provider(Provider_No),
	category_no	numeric_id
						CONSTRAINT Charges200406CategoryNoFK
							REFERENCES dbo.Category(Category_No),
	charge_dt		datetime 	NOT NULL
						CONSTRAINT Charges200406ChargeDtCK
							CHECK (Charge_dt >= '20040601' 
							        AND Charge_dt < '20040701'),
	charge_amt		money		NOT NULL,
	statement_no	numeric_id,
	charge_code	status_code
) ON ChargesStaging
go

INSERT Charges200406 (member_no, provider_no, category_no
						, charge_dt, charge_amt
						, statement_no, charge_code)
	SELECT member_no, provider_no, category_no
			, dateadd(yy, 5, charge_dt), (charge_amt + charge_no)/10
			, statement_no, charge_code 
	FROM CreditPV.dbo.Charge
	WHERE month(charge_dt) = 6
	ORDER BY charge_dt, charge_no
go

EXEC sp_help Charges200406
-- Will be on ChargesStaging
go

ALTER TABLE Charges200406
ADD CONSTRAINT Charges200406PK
		PRIMARY KEY CLUSTERED (charge_dt, charge_no) 
			ON [PartitionedViewsFilegroup1]
go

EXEC sp_help Charges200406
-- Will be on PartitionedViewsFilegroup1
go

----------------------------------------------------------------------------------------------
--------------------------------------- JULY --------------------------------------------
----------------------------------------------------------------------------------------------
CREATE TABLE Charges200407
( 	charge_no		numeric_id		NOT NULL	IDENTITY,
	member_no		numeric_id 	NOT NULL
						CONSTRAINT Charges200407MemberNoFK
							REFERENCES dbo.Member(Member_No),
	provider_no		numeric_id
						CONSTRAINT Charges200407ProviderNoFK
							REFERENCES dbo.Provider(Provider_No),
	category_no	numeric_id
						CONSTRAINT Charges200407CategoryNoFK
							REFERENCES dbo.Category(Category_No),
	charge_dt		datetime 	NOT NULL
						CONSTRAINT Charges200407ChargeDtCK
							CHECK (Charge_dt >= '20040701' AND Charge_dt < '20040801'),
	charge_amt		money		NOT NULL,
	statement_no	numeric_id,
	charge_code	status_code
) ON ChargesStaging
go

INSERT Charges200407 (member_no, provider_no, category_no
						, charge_dt, charge_amt
						, statement_no, charge_code)
	SELECT member_no, provider_no, category_no
			, dateadd(yy, 5, charge_dt), (charge_amt + charge_no)/10
			, statement_no, charge_code 
	FROM CreditPV.dbo.Charge
	WHERE month(charge_dt) = 7
	ORDER BY charge_dt, charge_no
go

EXEC sp_help Charges200407
-- Will be on ChargesStaging
go

ALTER TABLE Charges200407
ADD CONSTRAINT Charges200407PK
		PRIMARY KEY CLUSTERED (charge_dt, charge_no) 
			ON [PartitionedViewsFilegroup2]
go

EXEC sp_help Charges200407
-- Will be on PartitionedViewsFilegroup2
go

----------------------------------------------------------------------------------------------
--------------------------------------- August-------------------------------------------
----------------------------------------------------------------------------------------------
CREATE TABLE Charges200408
( 	charge_no		numeric_id		NOT NULL	IDENTITY,
	member_no		numeric_id 	NOT NULL
						CONSTRAINT Charges200408MemberNoFK
							REFERENCES dbo.Member(Member_No),
	provider_no		numeric_id
						CONSTRAINT Charges200408ProviderNoFK
							REFERENCES dbo.Provider(Provider_No),
	category_no	numeric_id
						CONSTRAINT Charges200408CategoryNoFK
							REFERENCES dbo.Category(Category_No),
	charge_dt		datetime 	NOT NULL
						CONSTRAINT Charges200408ChargeDtCK
							CHECK (Charge_dt >= '20040801' AND Charge_dt < '20040901'),
	charge_amt		money		NOT NULL,
	statement_no	numeric_id,
	charge_code	status_code
) ON ChargesStaging
go

INSERT Charges200408 (member_no, provider_no, category_no
						, charge_dt, charge_amt
						, statement_no, charge_code)
	SELECT member_no, provider_no, category_no
			, dateadd(yy, 5, charge_dt), (charge_amt + charge_no)/10
			, statement_no, charge_code 
	FROM CreditPV.dbo.Charge
	WHERE month(charge_dt) = 8
	ORDER BY charge_dt, charge_no
go

EXEC sp_help Charges200408
-- Will be on ChargesStaging
go

ALTER TABLE Charges200408
ADD CONSTRAINT Charges200408PK
		PRIMARY KEY CLUSTERED (charge_no) 
			ON [PartitionedViewsFilegroup3]
go

--ALTER TABLE Charges200408
--DROP CONSTRAINT Charges200408PK
--go

--ALTER TABLE Charges200408
--ADD CONSTRAINT Charges200408PK
--		PRIMARY KEY CLUSTERED (charge_dt, charge_no) 
--			ON [PartitionedViewsFilegroup3]
--go

EXEC sp_help Charges200408
-- Will be on PartitionedViewsFilegroup3
go

----------------------------------------------------------------------------------------------
--------------------------------------- Sept ----------------------------------------------
----------------------------------------------------------------------------------------------
CREATE TABLE Charges200409
( 	charge_no		numeric_id		NOT NULL	IDENTITY,
	member_no		numeric_id 	NOT NULL
						CONSTRAINT Charges200409MemberNoFK
							REFERENCES dbo.Member(Member_No),
	provider_no		numeric_id
						CONSTRAINT Charges200409ProviderNoFK
							REFERENCES dbo.Provider(Provider_No),
	category_no	numeric_id
						CONSTRAINT Charges200409CategoryNoFK
							REFERENCES dbo.Category(Category_No),
	charge_dt		datetime 	NOT NULL
						CONSTRAINT Charges200409ChargeDtCK
							CHECK (Charge_dt >= '20040901' AND Charge_dt < '20041001'),
	charge_amt		money		NOT NULL,
	statement_no	numeric_id,
	charge_code	status_code
) ON ChargesStaging
go

INSERT Charges200409 (member_no, provider_no, category_no
						, charge_dt, charge_amt
						, statement_no, charge_code)
	SELECT member_no, provider_no, category_no
			, dateadd(yy, 5, charge_dt), (charge_amt + charge_no)/10
			, statement_no, charge_code 
	FROM CreditPV.dbo.Charge
	WHERE month(charge_dt) = 9
	ORDER BY charge_dt, charge_no
go

EXEC sp_help Charges200409
-- Will be on ChargesStaging
go

ALTER TABLE Charges200409
ADD CONSTRAINT Charges200409PK
		PRIMARY KEY CLUSTERED (charge_no) 
			ON [PartitionedViewsFilegroup4]
go

--ALTER TABLE Charges200409
--DROP CONSTRAINT Charges200409PK
--go

--ALTER TABLE Charges200409
--ADD CONSTRAINT Charges200409PK
--		PRIMARY KEY CLUSTERED (charge_dt, charge_no) 
--			ON [PartitionedViewsFilegroup4]
--go

EXEC sp_help Charges200409
-- Will be on PartitionedViewsFilegroup4
go

----------------------------------------------------------------------------------------------
---------------------------------- Create View --------------------------------------
----------------------------------------------------------------------------------------------
CREATE VIEW dbo.ChargePV
AS
SELECT * FROM dbo.Charges200406
UNION ALL
SELECT * FROM dbo.Charges200407
UNION ALL
SELECT * FROM dbo.Charges200408
UNION ALL
SELECT * FROM dbo.Charges200409
go

----------------------------------------------------------------------------------------------
------------------------------ SET RECOVERY TO FULL and BACKUP! ------------------------------
----------------------------------------------------------------------------------------------
ALTER DATABASE [CreditPV]
	SET RECOVERY FULL
GO

BACKUP DATABASE [CreditPV]
TO DISK = N'C:\SQLskills\CreditPV.bak' 
WITH INIT, STATS = 10
GO

RESTORE HEADERONLY FROM DISK = N'C:\SQLskills\CreditPV.bak' 
GO

----------------------------------------------------------------------------------------------
------------------------------ Partition Elimination: query the view -------------------------
----------------------------------------------------------------------------------------------

SET STATISTICS IO ON
go

SELECT * 
FROM ChargePV
WHERE Charge_dt >= '20040714' 
	AND Charge_dt < '20040715' 
go

SELECT * 
FROM ChargePV
WHERE Charge_dt >= '20040914' 
	AND Charge_dt < '20040915' 
go

SELECT OBJECTPROPERTY(object_id('Charges200409ChargeDtCK'), 
	'CnstIsNotTrusted')
go

-- And, because of the PKs, 
--    we can't update through the view:
-- (1) they're not even consistent
-- (2) they don't all have the partitioning key

SELECT * 
FROM ChargePV 
WHERE charge_no = 146551 
    AND charge_dt = '20040712 10:45:23.027'
go

UPDATE ChargePV
    SET charge_amt = 1234
    where charge_no = 146551 
        AND charge_dt = '20040712 10:45:23.027'
go

--Msg 4445, Level 16, State 11, Line 1
--UNION ALL view 'CreditPV.dbo.ChargePV' is not updatable 
-- because the primary key of table '[CreditPV].[dbo].[Charges200408]' 
-- is not unioned with primary keys of preceding tables.

-- Or, if the keys don't have the PK then you'll get this:
--Msg 4436, Level 16, State 13, Line 1
--UNION ALL view 'CreditPV.dbo.ChargePV' is not updatable because a partitioning column was not found.



-- Go back and drop/recreate the constraints
BEGIN TRAN
UPDATE ChargePV
    SET charge_amt = 1234
    where charge_dt between '20040730' AND '20040802'
ROLLBACK     
go


UPDATE Charges200407
    SET charge_amt = 4567
    WHERE charge_no = 146551 
        AND charge_dt = '20040712 10:45:23.027'
go


-- PART 2: We'll discuss in Partial Database Availability 
--         and Online Piecemeal Restore

----------------------------------------------------------------------------------------------
---------------------------------- Disaster strikes! -----------------------------------------
----------------------------------------------------------------------------------------------

-- Pull out the USB device that has July's data:
SET STATISTICS IO ON
go

SELECT * 
FROM ChargePV
WHERE Charge_dt >= '20040914' 
	AND Charge_dt < '20040915' 
go

SELECT * 
FROM ChargePV
WHERE Charge_dt >= '20040702' 
	AND Charge_dt < '20040704' 
go

SELECT * 
FROM ChargePV
WHERE datename(DW, Charge_dt) = 'Monday'
go

-- So far - no problems... SQL Server hasn't even 
-- noticed that the drive has been removed...

-- Why?    The data's in cache!

-- We can even modify data:

SELECT * FROM ChargePV 
WHERE charge_no = 146551 
    AND charge_dt BETWEEN '20040701' AND '20040801'
go

UPDATE Charges200407
    SET charge_amt = 5678
    where charge_no = 146551 
        and charge_dt = '20040712 10:45:23.027'
go

SELECT * FROM ChargePV 
WHERE charge_no = 146551 
AND charge_dt BETWEEN '20040701' AND '20040801'
go

DBCC DROPCLEANBUFFERS
go

-- Let's check state first:
SELECT file_id, name, physical_name, state_desc 
FROM sys.database_files
go

ALTER DATABASE [CreditPV]
	MODIFY FILE (NAME = N'CreditPVFGFile2', OFFLINE)
GO

ALTER DATABASE [CreditPV]
	MODIFY FILE (NAME = N'CreditPVFGFile3', OFFLINE)
GO

----------------------------------------------------------------------------------------------
------------------------------ Restore the file ----------------------------------------------
----------------------------------------------------------------------------------------------

USE Master
GO

RESTORE DATABASE [CreditPV]
	FILE = N'CreditPVFGFile2'
	, FILE = N'CreditPVFGFile3'
FROM DISK = N'C:\SQLskills\CreditPV.bak' 
WITH FILE = 1, 
	MOVE N'CreditPVFGFile2' TO N'C:\SQLskills\CreditPVFGFile2.ndf', 
	MOVE N'CreditPVFGFile3' TO N'C:\SQLskills\CreditPVFGFile3.ndf', 
	NORECOVERY
go

-- Here, because PDA isn't supported on Standard Edition
-- We must backup the tail first and then we can start our
-- restore sequence - while the database is COMPLETELY offline

BACKUP LOG [CreditPV]
	TO DISK = N'C:\SQLskills\CreditPVLOG.bak' 
	WITH INIT, NO_TRUNCATE
GO

-- This command takes the database OFFLINE
-- However, we can still restore only the piece that's damaged.
-- This will reduce our total amount of downtime.
USE master
go

RESTORE DATABASE [CreditPV]
	FILE = N'CreditPVFGFile2'
	, FILE = N'CreditPVFGFile3'
FROM DISK = N'C:\SQLskills\CreditPV.bak' 
WITH FILE = 1, 
	MOVE N'CreditPVFGFile2' 
	    TO N'C:\SQLskills\CreditPVFGFile2.ndf', 
	MOVE N'CreditPVFGFile3' 
	    TO N'C:\SQLskills\CreditPVFGFile3.ndf', 
	NORECOVERY
go

-- Now restore the tail
RESTORE LOG [CreditPV]
FROM DISK = N'C:\SQLskills\CreditPVLOG.bak' 
	WITH FILE = 1, RECOVERY
go

-- Now, everyone can get back in!

-- From here, we could go on to some advanced scenarios as well
-- such as indexed views on the base tables of a partitioned view.
-- You CANNOT create an indexed view on a PV but if you create IVs
-- on the base tables, then SQL Server CAN use them. But, the
-- big problem when you're NOT on EE is that you can only use
-- them when referencing the IV directly. 

-- So, the point - you might still want to use PVs (even on EE) and 
-- when you do, then you can still get FANTASTIC benefits with
-- IVs!

-- The next section is an added benefit of Partitioning in that you can create indexed views
-- on the base tables and then have SQL Server aggregate the aggregates in a single query.

CREATE VIEW dbo.SumOfAllChargesByMemberJune04
	WITH SCHEMABINDING  -- required if you plan to index this view!
AS
SELECT c.member_no AS MemberNo, 
	COUNT_BIG(*) AS NumberOfCharges, -- required when GROUP BY is in an indexed view!
	SUM(c.charge_amt) AS TotalSales
FROM dbo.Charges200406 AS c
GROUP BY c.member_no
go

CREATE UNIQUE CLUSTERED INDEX SumofAllChargesIndex
	ON dbo.SumOfAllChargesByMemberJune04 (MemberNo) 
go

CREATE VIEW dbo.SumOfAllChargesByMemberJuly04
	WITH SCHEMABINDING  -- required if you plan to index this view!
AS
SELECT c.member_no AS MemberNo, 
	COUNT_BIG(*) AS NumberOfCharges, -- required when GROUP BY is in an indexed view!
	SUM(c.charge_amt) AS TotalSales
FROM dbo.Charges200407 AS c
GROUP BY c.member_no
go

CREATE UNIQUE CLUSTERED INDEX SumofAllChargesIndex
	ON dbo.SumOfAllChargesByMemberJuly04 (MemberNo) 
go

CREATE VIEW dbo.SumOfAllChargesByMemberAug04
	WITH SCHEMABINDING  -- required if you plan to index this view!
AS
SELECT c.member_no AS MemberNo, 
	COUNT_BIG(*) AS NumberOfCharges, -- required when GROUP BY is in an indexed view!
	SUM(c.charge_amt) AS TotalSales
FROM dbo.Charges200408 AS c
GROUP BY c.member_no
go

CREATE UNIQUE CLUSTERED INDEX SumofAllChargesIndex
	ON dbo.SumOfAllChargesByMemberAug04 (MemberNo) 
go

CREATE VIEW dbo.SumOfAllChargesByMemberSept04
	WITH SCHEMABINDING  -- required if you plan to index this view!
AS
SELECT c.member_no AS MemberNo, 
	COUNT_BIG(*) AS NumberOfCharges, -- required when GROUP BY is in an indexed view!
	SUM(c.charge_amt) AS TotalSales
FROM dbo.Charges200409 AS c
GROUP BY c.member_no
go

CREATE UNIQUE CLUSTERED INDEX SumofAllChargesIndex
	ON dbo.SumOfAllChargesByMemberSept04 (MemberNo) 
go

-- Aggregating the aggregates!
SELECT c.member_no AS MemberNo, 
	COUNT_BIG(*) AS NumberOfCharges, -- required when GROUP BY is in an indexed view!
	SUM(c.charge_amt) AS TotalSales
FROM ChargePV AS c
GROUP BY c.member_no
go

SELECT c.member_no AS MemberNo, 
	AVG(c.charge_amt) AS TotalSales
FROM ChargePV AS c
GROUP BY c.member_no
ORDER BY c.member_no
go

CREATE VIEW dbo.SumOfAllChargesPV
AS
SELECT * 
    FROM dbo.SumOfAllChargesByMemberJune04 
        WITH (NOEXPAND)
UNION ALL
SELECT * FROM dbo.SumOfAllChargesByMemberJuly04 WITH (NOEXPAND)
UNION ALL
SELECT * FROM dbo.SumOfAllChargesByMemberAug04 WITH (NOEXPAND)
UNION ALL
SELECT * FROM dbo.SumOfAllChargesByMemberSept04 WITH (NOEXPAND)
go

-- On EE then even this query would work:
SELECT c.member_no AS MemberNo, 
	AVG(c.charge_amt) AS TotalSales
FROM ChargePV AS c
GROUP BY c.member_no
ORDER BY c.member_no
go

-- On All other editions you MUST use the PV where
-- the IVs are UNIONed (UNION ALL) and where they
-- use the NOEXPAND hint
SELECT c.memberNo AS MemberNo, 
	AVG(c.TotalSales) AS TotalSales
FROM SumOfAllChargesPV AS c
GROUP BY c.memberNo
ORDER BY c.memberNo
go
